#import "CoreDataStack.h"
#import "Constants.h"
#import "FolderWatcher.h"
@import UIKit;

@interface CoreDataStack()
@property (nonatomic, readwrite, strong) NSManagedObjectContext *context;
@property (nonatomic, strong) FolderWatcher *folderWatcher;
@end

@implementation CoreDataStack

- (instancetype)initWithStorageURL:(NSURL*)storageURL
{
    if (!(self = [super init])) return nil;
    
    [self setUpCoreDataStackWithStorageURL:storageURL];
    return self;
}

- (void)setUpCoreDataStackWithStorageURL:(NSURL*)storageURL
{
    NSManagedObjectModel *model = [NSManagedObjectModel mergedModelFromBundles:[NSBundle allBundles]];
    NSPersistentStoreCoordinator *psc = [[NSPersistentStoreCoordinator alloc] initWithManagedObjectModel:model];
    NSDictionary *options = @{
      NSPersistentStoreFileProtectionKey: NSFileProtectionComplete,
      NSMigratePersistentStoresAutomaticallyOption: @YES
    };
    NSError *error = nil;
    NSPersistentStore *store = [psc addPersistentStoreWithType:NSSQLiteStoreType configuration:nil URL:storageURL options:options error:&error];
    
    if (!store) {
        NSLog(@"Error adding persistent store: %@",error);
        abort();
    }
    
    self.context = [[NSManagedObjectContext alloc] initWithConcurrencyType:NSMainQueueConcurrencyType];
    self.context.persistentStoreCoordinator = psc;
}

- (void)save
{
    [self.context performBlockAndWait:^{
        NSError *error = nil;
        [self.context save:&error];
        if (error) {
            NSLog(@"Failed to save: %@",error);
        }
    }];
}

#pragma mark - Sending updates to other apps in the app group

-(NSURL*)tickleURL
{
    NSURL *tickleURL = [[NSFileManager defaultManager] containerURLForSecurityApplicationGroupIdentifier:AppGroupIdentifier];
    tickleURL = [tickleURL URLByAppendingPathComponent:TickleFolderPath isDirectory:YES];
    return tickleURL;
}

-(void)setSendsUpdates:(BOOL)sendsUpdates
{
    if (sendsUpdates == _sendsUpdates) return;
    
    if (sendsUpdates) {
        [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(sendUpdatesFromContextSaved:) name:NSManagedObjectContextDidSaveNotification object:self.context];
    } else {
        [[NSNotificationCenter defaultCenter] removeObserver:self name:NSManagedObjectContextDidSaveNotification object:self.context];
    }
    
    _sendsUpdates = sendsUpdates;
}

-(void)sendUpdatesFromContextSaved:(NSNotification*)saveNotification
{
    NSURL *tickleURL = [self tickleURL];
    [[NSFileManager defaultManager] createDirectoryAtURL:tickleURL withIntermediateDirectories:YES attributes:nil error:nil];
    NSString *filename = [[NSUUID UUID] UUIDString];
    tickleURL = [tickleURL URLByAppendingPathComponent:filename];
    
    NSDictionary *saveInfo = [self serializableDictionaryFromSaveNotification:saveNotification];
    
    [saveInfo writeToURL:tickleURL atomically:YES];
}

- (NSDictionary*)serializableDictionaryFromSaveNotification:(NSNotification*)saveNotification
{
    NSMutableDictionary *saveInfo = [NSMutableDictionary dictionary];
    for (NSString *key in @[NSInsertedObjectsKey,NSDeletedObjectsKey,NSUpdatedObjectsKey]) {
        NSArray *managedObjects = saveNotification.userInfo[key];
        if (!managedObjects) continue;
        NSMutableArray *objectIDRepresentations = [NSMutableArray array];
        for (NSManagedObject *object in managedObjects) {
            NSManagedObjectID *objectID = [object objectID];
            NSURL *URIRepresentation = [objectID URIRepresentation];
            NSString *objectIDValue = [URIRepresentation absoluteString];
            [objectIDRepresentations addObject:objectIDValue];
        }
        saveInfo[key] = [objectIDRepresentations copy];
    }
    return [saveInfo copy];
}

#pragma mark - Receiving updates

- (void)setReceivesUpdates:(BOOL)receivesUpdates
{
    if (receivesUpdates == _receivesUpdates) return;
    
    if (receivesUpdates) {
        [self clearTickleFolder];
        NSURL *tickleURL = [self tickleURL];
        [[NSFileManager defaultManager] createDirectoryAtURL:tickleURL withIntermediateDirectories:YES attributes:nil error:nil];
        __weak typeof(self) weakSelf = self;
        self.folderWatcher = [[FolderWatcher alloc] initWithFolderURL:tickleURL writeAction:^{
            [weakSelf updateWrittenToURL:tickleURL];
        }];
        [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(clearTickleFolder) name:UIApplicationWillEnterForegroundNotification object:nil];
    } else {
        self.folderWatcher = nil;
        [[NSNotificationCenter defaultCenter] removeObserver:self name:UIApplicationWillEnterForegroundNotification object:nil];
    }
    
    _receivesUpdates = receivesUpdates;
}

- (void)updateWrittenToURL:(NSURL*)url
{
    NSError *error = nil;
    NSFileManager *fileManager = [NSFileManager defaultManager];
    NSArray *files = [fileManager contentsOfDirectoryAtURL:url includingPropertiesForKeys:@[NSURLAddedToDirectoryDateKey] options:0 error:&error];
    if (error) NSLog(@"Error getting directory contents %@",error);
    
    if ([files count] == 0) return;
    
    NSArray *sortedFiles = [files sortedArrayUsingComparator:^NSComparisonResult(NSURL* obj1, NSURL* obj2) {
        
        NSDate *added1 = nil, *added2 = nil;
        NSError *error1 = nil, *error2 = nil;
        
        BOOL extracted1 = [obj1 getResourceValue:&added1 forKey:NSURLAddedToDirectoryDateKey error:&error1];
        BOOL extracted2 = [obj2 getResourceValue:&added2 forKey:NSURLAddedToDirectoryDateKey error:&error2];
        if (extracted1 && extracted2) {
            return [added1 compare:added2];
        } else {
            NSLog(@"Error extracting: %@ and/or %@",error1,error2);
            return NSOrderedSame;
        }
    }];
    
    [self.context performBlockAndWait:^{
        NSPersistentStoreCoordinator *coordinator = self.context.persistentStoreCoordinator;
        
        for (NSURL *fileURL in sortedFiles) {
            NSDictionary *updateDictionary = [[NSDictionary alloc] initWithContentsOfURL:fileURL];
            
            NSNotification *importNotification = [self importNotificationFromDictionary:updateDictionary coordinator:coordinator];
            [self.context mergeChangesFromContextDidSaveNotification:importNotification];
            NSError *deleteError = nil;
            [fileManager removeItemAtURL:fileURL error:&deleteError];
            if (deleteError) NSLog(@"Error removing %@ : %@",fileURL,deleteError);
        }
    }];
}

- (NSNotification*)importNotificationFromDictionary:(NSDictionary*)updateDictionary coordinator:(NSPersistentStoreCoordinator*)coordinator
{
    NSMutableDictionary *userInfo = [NSMutableDictionary dictionary];
    for (NSString *key in [updateDictionary allKeys]) {
        NSMutableArray *newObjectIDs = [NSMutableArray array];
        for (NSString *objectID in updateDictionary[key]) {
            NSURL *URIRepresentation = [NSURL URLWithString:objectID];
            NSManagedObjectID *managedObjectID = [coordinator managedObjectIDForURIRepresentation:URIRepresentation];
            [newObjectIDs addObject:managedObjectID];
        }
        userInfo[key] = newObjectIDs;
    }
    
    NSNotification *importNotification = [NSNotification notificationWithName:NSPersistentStoreDidImportUbiquitousContentChangesNotification object:coordinator userInfo:userInfo];
    return importNotification;
}

- (void)clearTickleFolder
{
    NSFileManager *fileManager = [NSFileManager defaultManager];
    NSArray *files = [fileManager contentsOfDirectoryAtURL:[self tickleURL] includingPropertiesForKeys:nil options:0 error:nil];
    for (NSURL *fileURL in files) {
        NSError *deleteError = nil;
        [fileManager removeItemAtURL:fileURL error:&deleteError];
        if (deleteError) NSLog(@"Error when clearing folder %@: %@",fileURL,deleteError);
    }
}

@end
